/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/

#include "common.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#if !defined(WIN32) && !defined(Q_OS_WIN)
#ifdef USE_DEBUG_TIMING
#include <sys/time.h>
#endif
#endif

unsigned int g_start_timer;
unsigned int getStartMillisecs(const char *message)
{
#if !defined(WIN32) && !defined(Q_OS_WIN)
#ifdef USE_DEBUG_TIMING
    struct timeval time;
    gettimeofday(&time, NULL);
    printf("Start timing for %s\n",message);
    g_start_timer=(time.tv_sec * 1000) + (time.tv_usec / 1000);
    return g_start_timer;
#else
    return 0;
#endif
#else
	return 0;
#endif
}

unsigned int getEndMillisecs()
{
#if !defined(WIN32) && !defined(Q_OS_WIN)
#ifdef USE_DEBUG_TIMING
    struct timeval time;
    gettimeofday(&time, NULL);
    unsigned int end_time=((time.tv_sec * 1000) + (time.tv_usec / 1000))-g_start_timer;
    printf("Elapsed time : %i\n",end_time);
    return end_time;
#else
    return 0;
#endif
#else
	return 0;
#endif
}
void compile_box(JPXT_SUPERBOX *super_box,unsigned int payload_size,unsigned int TBox)
{
    memset(super_box,0,sizeof(JPXT_SUPERBOX));
    XDP_SWAP_SHORT16(super_box->jpg_tag,0xFFEB);
    XDP_SWAP_SHORT16(super_box->Le,(sizeof(JPXT_SUPERBOX)-2)+payload_size);
    XDP_SWAP_SHORT16(super_box->CI,0x4A50); // 'JP'
    XDP_SWAP_SHORT16(super_box->En,1);
    XDP_SWAP_SHORT16(super_box->Z,1);
    XDP_SWAP_SHORT32(super_box->LBox,payload_size+sizeof(super_box->LBox)+sizeof(super_box->TBox));
    XDP_SWAP_SHORT32(super_box->TBox,TBox);
}

float swap_float(float value)
{
    float out_value;
    unsigned char *o=(unsigned char *)&value;
    unsigned char *p=(unsigned char *)&out_value;

    p[0]=o[3];
    p[1]=o[2];
    p[2]=o[1];
    p[3]=o[0];

    return out_value;
}

float toFloat(unsigned char in_byte, int Rw, int e)
{
    return in_byte/(powf(2,(float)Rw)-e);
}

unsigned char toByte(float in_value,int Rw, int e)
{
    return in_value*(powf(2,(float)Rw)-e);
}
