/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/


#ifndef _jpeg_io_h
#define _jpeg_io_h
#include "../jpeg/jpeglib.h"
#include "../jpeg/jerror.h"
#include <setjmp.h>
#include "image.h"

struct my_error_mgr {
    struct jpeg_error_mgr pub;	/* "public" fields */
    
    jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;
#define OUTPUT_BUF_SIZE 4096
typedef struct tagDestinationManager {
	/// public fields
	struct jpeg_destination_mgr pub;
	/// destination stream
	FILE *outfile;
    JOCTET *out_buffer;
    unsigned int out_buffer_size;
	/// start of buffer
	JOCTET * buffer;
} DestinationManager;
typedef DestinationManager* xdpjpg_dst_ptr;

enum {
    JPEG_SUBSAMPLE_DEFAULT,
    JPEG_SUBSAMPLE_422,
    JPEG_SUBSAMPLE_411,
    JPEG_SUBSAMPLE_444
};

typedef struct _jpg_info
{
    int dct_type;
    int optimized_huffman;
    int lossless;
    int compression_level;
} JPG_INFO;

int jpeg_getSize(const char *filename, int *width, int *height, int *grayscale);
FILE_IMAGE *jpeg_load ( const char *filename ); // Load Jpeg file from file
FILE_IMAGE *jpeg_load_mem ( MEMORY_FILE_IMAGE *image ); // Load Jpeg file from memory
FILE_IMAGE *jpeg_load_mem_p ( unsigned char *data, size_t data_size);
int jpeg_write(FILE_IMAGE *image, const char *filename, int quality,int subsample,int dct_opt,int huffman_opt, int lossless);   // Save Jpeg file on disk
MEMORY_FILE_IMAGE *jpeg_write_mem(FILE_IMAGE *image, int quality,int subsample,int dct_opt,int huffman_opt, int lossless); // Save Jpeg file into memory
unsigned short computeChecksum(MEMORY_FILE_IMAGE *jpeg_image);
#endif
