/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/

#include "tone_mapping.h"

FILE_IMAGE *tone_mapping(int tmo,FILE_IMAGE *image,float gamma)
{
    FILE_IMAGE *out=NULL;
    if (tmo==CLAMP_TMO)
        out=gamma_tm(image,gamma);
    else if (tmo==REINHARD02_TMO)
        out=reinhard02(image,gamma);

    return out;
}

FILE_IMAGE *reinhard02(FILE_IMAGE *image,float gamma)
{
    int width=image->width;
    int height=image->height;
    unsigned input_channels=3;
	FILE_IMAGE *out = NULL;
	unsigned pitch;
	unsigned char *output;
	int x, y;
    float red, green, blue;
    float sum = 0, Lw_bar, Y;
    float L, Lw, Ld;
    float max = 0, scale;

    const float *input=(const float *)(image->data);
    if (image->grayscale==1)
        input_channels=1;
    out=allocate_image(image->width,image->height,96,input_channels);
    output=(unsigned char*)out->data;
    pitch=(width*input_channels)*sizeof(float);



    // todo: reinhard tonemapping
    for(y = 0; y < height; ++y)
        for(x = 0; x < width; ++x){
            const int offset = (x + y * width) * input_channels;
            const float* src = input + offset;

            red= src[0]; // R
            if (input_channels==3)
            {
                green = src[1]; // G
                blue = src[2]; // B

                Y = red * 0.2126 + green * 0.7152 + blue * 0.0722;
            }
            else
                Y=red;

            if(Y > max) max = Y;
            sum += ( log(0.0001+Y) );
        }
    sum /= (height*width);
    Lw_bar = exp(sum);
    //printf("Lw = %f\n", Lw_bar);

    for(y = 0; y < height; ++y) {
        float *dst=(float*)output;
        unsigned int off=0;
        for(x = 0; x < width; ++x){
            const int offset = (x + y * width) * input_channels;
            const float* src = input + offset;
            red= src[0]; // R

            if (input_channels==3)
            {
                green = src[1]; // G
                blue = src[2]; // B


                Y = red * 0.2126 + green * 0.7152 + blue * 0.0722;
            }
            else
                Y=red;

            Lw = Y;
            L = 0.18 * Lw / Lw_bar;
            Ld = L / (1 + L);

            scale = Ld / Y;

            red *= scale;
            if (input_channels==3)
            {
                green *= scale;
                blue *= scale;
            }
            red=red>1.0f?1.0f:red;
            red=red<0.0f?0.0f:red;

            red=pow(red,1.0/gamma);

            if (input_channels==3)
            {
                green=green>1.0f?1.0f:green;
                blue=blue>1.0f?1.0f:blue;
                green=green<0.0f?0.0f:green;
                blue=blue<0.0f?0.0f:blue;
                green=pow(green,1.0/gamma);
                blue=pow(blue,1.0/gamma);

                dst[off+1] = green; // G
                dst[off+2] = blue; // R
                dst[off] = red; // B
                off+=3;
            }
            else
            {
                dst[off] = red; // B
                off++;
            }

        }
        output += pitch;
        }
    return out;
}

FILE_IMAGE *gamma_tm(FILE_IMAGE *image, float gamma)
{
    int x, y;
    float red, green, blue;
    int width=image->width;
    int height=image->height;
	const float *input;
	FILE_IMAGE *out;
	unsigned char *output;
	unsigned pitch;
    unsigned input_channels=3;
    if (image->grayscale==1)
        input_channels=1;
    input=(const float *)(image->data);
    out=allocate_image(image->width,image->height,96,input_channels);
    output=(unsigned char*)out->data;
    pitch=(width*input_channels)*sizeof(float);
    for(y = 0; y < height; ++y) {
        float *dst=(float*)output;
        unsigned int off=0;
        for(x = 0; x < width; ++x){
            const int offset = (x + y * width) * input_channels;
            const float* src = input + offset;

            red= src[0]; // R

            red=red>1.0f?1.0f:red;
            red=red<0.0f?0.0f:red;

            red=pow(red,1.0/gamma);

            if (input_channels==3)
            {
                green = src[1]; // G
                blue = src[2]; // B

                green=green>1.0f?1.0f:green;
                blue=blue>1.0f?1.0f:blue;
                green=green<0.0f?0.0f:green;
                blue=blue<0.0f?0.0f:blue;

                green=pow(green,1.0/gamma);
                blue=pow(blue,1.0/gamma);

                dst[off+2] = blue; // R
                dst[off+1] = green; // G
                dst[off] = red; // B
            }
            else
                dst[off] = red;

            off+=input_channels;
        }
        output += pitch;
    }

    return out;
}
