/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jpeg_io.h"
#include "hdr_io.h"
#include "image.h"
#include "png_io.h"
#include "pfm_io.h"
#include "ppm_io.h"
#include "encoder.h"
#include "xdp_jpeg_hdr.h"
#include "tone_mapping.h"
#include "common.h"

char out_filename[512]="output.jpg";
char out_hdr_filename[]="output.hdr";
int g_jpeg_dct=JDCT_ISLOW;
int g_tmo=CLAMP_TMO;
int g_Rw=8;
int g_e=1;
int g_huffman_opt=1;
int g_lossless=0;
int g_input_shift=1;
int g_log_decode=0;
int g_save_resi_bitstream=0;
float g_ocon_p1=0;
float g_ocon_p2=0;
float g_ocon_p3=1.0f;
float g_ocon_p4=0;
int g_use_custom_ocon=0;

void print_help_enc()
{
    printf("\nXDepth-HDR Encoder options\n\n");
    printf("Use :\n xdp_encoder [operation] [optsions] <input_hdr_file> <output_jpeg_xt_file>\n");
    printf("Options:\n");
    printf(" -q : Ldr quality (0-100) (default 97)\n -Q : Hdr quality (0-100) (default 100)\n -quality : Global image quality (0-200)\n -ls : Ldr subsample (default 2 - 411)\n -hs : Hdr subsample (default 3 - 444)\n");
    printf(" -ldr : Tone mapped image in PPM format or PNG (with libpng installed) format (8-bit per channel no alpha)\n");
    printf(" -xyz : XYZ HDR image expected in input. (default disabled)\n -resi_no_xyz : Do not apply XYZ transformation on residual data. (default disabled)\n");
    printf(" -srgb : Use srgb profile to linearize tone mapped image. (default disabled)\n");
    printf(" -dct : 0 = accurate integer, 1 = less accurate integer, 2 = accurate float (default 0)\n");
    printf(" -dct : 0 = accurate integer, 1 = less accurate integer, 2 = accurate float (default 0)\n");
    printf(" -tmo : 0 = clamp tmo, 1 = reinhard02 (default 0)\n");
    printf(" -lg : Ldr gamma to linearize tone mapped image. (default 1.0)\n");
    printf(" -rcenter : Enable centered rounding mode. (default keep extremes)\n");
    printf(" -re : Save 'residual.jpg' file, the residual bitstream.\n");
    printf(" -no_huff : Disable Huffman optimization on base and residual layer. (default enabled)\n\n");
    printf("Reserved for special experiments.\n");
    printf(" Specify parameters for Output Conversion Box.\n");
    printf(" -p1 <value> -p2 <value> -p3 <value> -p4 <value>\n\n");
    printf("Sub sample option values:\n     0 = value 411\n     1 = 422\n     2 = 411\n     3 = 444 (3 = Default value)\n\n");
    printf("Input file format supported is HDR (Radiance) or PFM.\n");
    printf("Example:\nTo encode inputFile.hdr in output.jpg with LDR quality of 100 and Residual quality of 100 with a subsample of 444 use the following:\n");
    printf ("  ./xdp_encdoder e -ldr tonemapped_image.ppm -q 100 -Q 100 -ls 3 -hs 3 inputFile.hdr output.jpg\n\n");
    printf("\nTo encode inputFile.pfm in output.jpg with default values and linearize tone mapped image using srgb profile use :\n");
    printf ("  ./xdp_encdoder e -ldr tonemapped_image.ppm -srgb inputFile.pfm output.jpg\n\n");
    printf("\nTo encode inputFile.pfm in output.jpg with default values and linearize tone mapped image using gamma value :\n");
    printf ("  ./xdp_encdoder e -ldr tonemapped_image.ppm -lg 2.2 inputFile.pfm output.jpg\n\n");
    printf("\nTo encode InputFile.pfm in output.jpg with default values with reinhard tone mapping :\n");
    printf ("  ./xdp_encdoder e -tmo 1 inputFile.pfm output.jpg\n\n");
}

void print_help_dec()
{
    printf("\nXDepth-HDR Decoder options\n\n");
    printf("Use :\n xdp_decoder [operation] <jpeg_xt_file> <output_hdr_file>\n\n");
    printf("Output file format supported are HDR or PFM files depending on output file extension.\n\n");
}

void print_help_main()
{
    printf("\nXDepth-HDR Encoder/Decoder Software Reference V.0.8, Copyright (C) 2012 Trellis Management Co. Ltd.\n\n");
    printf("Please specify the operation:\n  e : encode JPEG-XT image\n  d : decode JPEG-XT image\n\n");
    print_help_enc();
    print_help_dec();
}

int main(int argc, char *argv[])
{
    int encode=0;
    char *output_file=out_filename;
    char *input_file=NULL;
    char *ldr_image=NULL;
    int hdr_quality=100;
    int ldr_quality=97;
    int hdr_subsamp=3;
    int ldr_subsamp=2;
    int quality=0;
    int srgb=0;
    int input_xyz=0;
    int no_resi_xyz=0;
    float lg=1.0f;
    float hg=1.0f;
    int automatic_hdr_gamma=1;
    int use_prencode=0;
    FILE_IMAGE *current_image=NULL;
    FILE_IMAGE *input_tm_file=NULL;
    FILE *tmpf;
	FILE_IMAGE *image;

    int ret;
    int c;
    float base_matrix[9]={g_XYZ2RGB[0][0],g_XYZ2RGB[0][1], g_XYZ2RGB[0][2],
                          g_XYZ2RGB[1][0],  g_XYZ2RGB[1][1],  g_XYZ2RGB[1][2],
                          g_XYZ2RGB[2][0], g_XYZ2RGB[2][1],  g_XYZ2RGB[2][2]};
    float resi_matrix[9]={g_XYZ2RGB[0][0],g_XYZ2RGB[0][1], g_XYZ2RGB[0][2],
                          g_XYZ2RGB[1][0],  g_XYZ2RGB[1][1],  g_XYZ2RGB[1][2],
                          g_XYZ2RGB[2][0], g_XYZ2RGB[2][1],  g_XYZ2RGB[2][2]};


    g_ocon_p2=swap_float(1.0f/(powf(2.0f,8.0f+8.0f) -1.0f));
    if (argc>1)
    {
        if (!strcmp(argv[1],"e"))
            encode=1;
        else if (!strcmp(argv[1],"d"))
            encode=0;
        else
        {
            print_help_main();
            return 0;
        }

        if (argc>2)
        {
            if (encode==1)
            {
                for ( c=2;c<=(argc-1);c++)
                {
                    if (strcmp(argv[c],"-quality")==0)
                    {
                        quality=1;
                        hdr_quality=atoi(argv[c+1])*0.5f;
                        ldr_quality=hdr_quality-(hdr_quality*0.03);
                        c++;
                    }
                    else if (strcmp(argv[c],"-q")==0)
                    {
                        if (quality==0)
                        {
                            ldr_quality=atoi(argv[c+1]);
                            c++;
                        }
                    }
                    else if (strcmp(argv[c],"-Q")==0)
                    {
                        if (quality==0)
                        {
                            hdr_quality=atoi(argv[c+1]);
                            c++;
                        }
                    }
                    else if (strcmp(argv[c],"-ls")==0)
                    {
                        ldr_subsamp=atoi(argv[c+1]);
                        c++;
                    }
                    else if (strcmp(argv[c],"-hs")==0)
                    {
                        hdr_subsamp=atoi(argv[c+1]);
                        c++;
                    }
                    else if (strcmp(argv[c],"-pe")==0)
                    {

                    }
                    else if (strcmp(argv[c],"-re")==0)
                    {
                        g_save_resi_bitstream=1;
                    }
                    else if (strcmp(argv[c],"-ldr")==0)
                    {
                        ldr_image=argv[c+1];
                        c++;
                    }

                    else if (strcmp(argv[c],"-dct")==0)
                    {
                        int v=atoi(argv[c+1]);
                        g_jpeg_dct=v==0?JDCT_ISLOW:g_jpeg_dct;
                        g_jpeg_dct=v==1?JDCT_IFAST:g_jpeg_dct;
                        g_jpeg_dct=v==2?JDCT_FLOAT:g_jpeg_dct;
                        c++;
                    }
                    else if (strcmp(argv[c],"-tmo")==0)
                    {
                        g_tmo = atoi(argv[c+1]);
                        c++;
                    }
                    else if (strcmp(argv[c],"-xyz")==0)
                    {
                        input_xyz=1;
                        printf("XYZ transformation enabled.\n");
                    }
                    else if (strcmp(argv[c],"-no_huff")==0)
                    {
                        g_huffman_opt=0;
                        printf("Huffman optimizations disabled.\n");
                    }
                    else if (strcmp(argv[c],"-rcenter")==0)
                    {
                        printf("Enable centered rounding mode.\n");
                        g_e=0;
                    }
                    else if (strcmp(argv[c],"-lss")==0)
                    {
                    }
                    else if (strcmp(argv[c],"-resi_no_xyz")==0)
                    {
                        printf ("No XYZ transformation on residual enabled.\n");
                        no_resi_xyz=1;                        
                        resi_matrix[0]=1.0f;
                        resi_matrix[1]=0.0f;
                        resi_matrix[2]=0.0f;
                        resi_matrix[3]=0.0f;
                        resi_matrix[4]=1.0f;
                        resi_matrix[5]=0.0f;
                        resi_matrix[6]=0.0f;
                        resi_matrix[7]=0.0f;
                        resi_matrix[8]=1.0f;
                    }
                    else if (strcmp(argv[c],"-lg")==0)
                    {
                        lg=(float) atof(argv[c+1]);
                        c++;
                    }
                    else if (strcmp(argv[c],"-hg")==0)
                    {
                        hg=(float) atof(argv[c+1]);
                        automatic_hdr_gamma=0;
                        c++;
                    }
                    else if (strcmp(argv[c],"-srgb")==0)
                    {
                        printf ("SRGB profile enabled.\n");
                        srgb=1;
                    }
                    else if (strcmp(argv[c],"-p1")==0)
                    {
                        g_ocon_p1=atof(argv[c+1]);
                        c++;
                    }
                    else if (strcmp(argv[c],"-p2")==0)
                    {
                        g_ocon_p2=atof(argv[c+1]);
                        c++;
                    }
                    else if (strcmp(argv[c],"-p3")==0)
                    {
                        g_ocon_p3=atof(argv[c+1]);
                        g_use_custom_ocon=1;
                        c++;
                    }
                    else if (strcmp(argv[c],"-p4")==0)
                    {
                        g_ocon_p4=atof(argv[c+1]);
                        c++;
                    }
                    else
                    {
                        if (input_file==NULL)
                        {
                            tmpf=fopen(argv[c],"rb");
                            if (tmpf!=NULL)
                                fclose(tmpf);
                            else
                            {
                                printf("Error: input file does not exists!\n");
                                return 0;
                            }

                            input_file=argv[c];
                        }
                        else
                            output_file=argv[c];
                    }
                }
                if (ldr_quality>100)
                    ldr_quality=100;
                if (ldr_quality<0)
                    ldr_quality=0;
                if (hdr_quality>100)
                    hdr_quality=100;
                if (hdr_quality<0)
                    hdr_quality=0;
                if (ldr_subsamp>3)
                    ldr_subsamp=3;
                if (ldr_subsamp<0)
                    ldr_subsamp=0;
                if (hdr_subsamp>3)
                    hdr_subsamp=3;
                if (hdr_subsamp<0)
                    hdr_subsamp=0;
                if (input_file==NULL)
                {
                    printf("No HDR or PFM input file specified!\n");
                    return 0;
                }
                START_TIMER("Start loading files");
                if (ldr_image!=NULL)
                {
                    printf("Base image : %s\n",ldr_image);
#ifndef DONT_USE_PNG
                    if (!strcmp(get_extension(ldr_image),"png"))
                    {
                        input_tm_file=png_load(ldr_image);
                        flip_vertical(input_tm_file);
                    }
#endif
                    if (!strcmp(get_extension(ldr_image),"ppm"))
                        input_tm_file=ppm_load(ldr_image);

                    if (input_tm_file==NULL)
                    {
                        printf("Error loading tone mapped file. Invalid format or image does not exists.\n");
                        return 0;
                    }

                }

                
                if (!strcmp(get_extension(input_file),"hdr") || !strcmp(get_extension(input_file),"rgbe"))
                {

                    printf("Original HDR file : %s\n",input_file);
                    image=hdr_load(input_file);
                    if (image == NULL)
                    {
                        printf("Error loading HDR file: %s\n",input_file);
                        return 0;
                    }
                    current_image=image;
                }
                else if (!strcmp(get_extension(input_file),"pfm"))
                {
                    printf("original PFM file : %s\n",input_file);
                    image=pfm_load(input_file);
                    if (image == NULL)
                    {
                        printf("Error loading PFM file: %s\n",input_file);
                        return 0;
                    }
                    current_image=image;
                }
                else
                {
                    printf("Input file format not supported, please use HDR file or PFM file!\n");
                    return 0;
                }
                if (ldr_image!=NULL)
                {
                if (input_tm_file->width != current_image->width || input_tm_file->height != current_image->height)
                {
                    printf ("Input Base image size differ from input HDR image.\n");
                    printf("Base image size : %ix%i\n",input_tm_file->width,input_tm_file->height);
                    printf("HDRimage size : %ix%i\n",current_image->width,current_image->height);
                    return 0;
                }
                }
                END_TIMER();
                printf("Image size : %ix%i\n",current_image->width,current_image->height);
                printf("Start encoding in JPeg-XT (profile B).\n");

                ret=xdp_encode(current_image,use_prencode,output_file,1.0f,hdr_quality,hdr_subsamp,ldr_quality,ldr_subsamp,lg,srgb,hg,automatic_hdr_gamma,input_tm_file,input_xyz,base_matrix,resi_matrix,no_resi_xyz);

                if (!ret)
                    printf("Error during encoding operation.\n");
                else
                    printf("File encoded successfully!\n");
            }
            else
            {
                for ( c=2;c<=(argc-1);c++)
                {

                    if (strcmp(argv[c],"-log")==0)
                    {
                        printf("Enable LOG decoding.\n");
                        g_log_decode=1;
                    }
                    else
                    {
                        input_file=argv[c];
                        //if (argc>2)
                             output_file=argv[c+1];
                        //else
                        //    output_file=out_hdr_filename;
                        break;
                    }
                }
                printf("Decoding JPeg-XT (profile B) image file.\n");
                START_TIMER("Start dencoding");
                ret = xdp_decode(input_file,output_file);
                END_TIMER();
                if (ret<1)
                {
                    if (ret==XT_ERROR_CHECKSUM)
                        printf("Checksum error!! Error decoding JPeg-XT image.\n");
                    else if (ret==XT_ERROR_INVALID_PROFILE)
                        printf("Invalid JPeg-XT profile, expected Profile-B.\n");
                    else
                        printf("Error decoding JPeg-XT image.\n");
                }
                else
                    printf("JPeg-XT image file decoded successfully!\n");

            }
        }
        else
        {
            if (encode==1)
                print_help_enc();
            else
                print_help_dec();
            return 0;
        }
    }
    else
    {
        print_help_main();
    }
     return 0;
}

